#!/usr/bin/perl -w
use Carp;
#use IO::File;
#no Carp 'verbose';
#
# utility functions used by several scripts
# utilities_pruned.pl was created by deleting many subroutines from "utilities.pl", leaving those
#  that are (or might be) used in the rust models; so its functions are a subset of those in "utilities"
#
######################################################################
#
#  subroutine declarations and prototypes
#
######################################################################

sub log10 ( $ );
sub dumphash ( $% );
sub setyear( $ );
sub is_a_number( $ );
sub datestr( $$ );
sub datetodoywinter( $$ );
sub datetodoy( $$ );
sub doytodatewinter( $ );
sub doytodate( $ );
sub make_doy( $ ); # convert date to doy if needed
sub make_doywinter( $ );
sub make_date( $ ); # convert date to doy if needed
sub incdate( \$\$ );
sub incdate2( $$ );
sub incdate_str( $ );
sub decdate2( $$ );
sub cmpdatenum ( $$$$ );
sub cmpdatestr;
sub cmpdatestr_ab;
sub max( $$ );
sub min( $$ );
sub trim_white( $ );
sub polynomial( $@ );
sub hhmm_hrsmins ( $ );
sub hhmm_hrs ( $ );
sub hhmm_mins ( $ );
sub hhmm ( $$ );
sub decimal_hrs ( $$ );
sub hhmm_diff( $$ );
sub dms2dd( $ );
sub dd2dms( $;$ );
sub dms( $;$ );
sub dd( $ );
sub round_dms( $ );



######################################################################
#
#  general time-related variables
#
######################################################################

# for timelocal(), inverse of localtime();
use Time::Local;
# for anonymous, self-closing filehandles
use IO::File;

# indices into the array returned by localtime()
$YEAR = 5; # years since 1900
$MON = 4;
$DAY = 3;

# Set up variables used in timestamps and for defining today
# PROBLEM: There needs to be a clear policy about these variables.  Are
# immutable after initialization, and guaranteed to refer to the present?
# In that case, they should either have less common names -- probably going
# in a packages.  Or are they to be overwritten with the current target
# date/time.  Right now it's a mess because $year and $year4 are
# overwritten in this way, but the others are not.  Maybe a subroutine
# should be free to set them for its desendents, but not for its siblings
# and ancestors -- this could probably be done with local(), or it might
# require a manually-manipulated stack, or the vars in a hash or somesuch.
@time = localtime(time);
# @time now contains these fields. Note that month is 0-11, not 1-12.
# sec, min, hour, day, month, year, day-of-week, day-of-year, isdst (?)

# use sprintf("%02d", ...) to convert number to 2-character, 0-padded string
#$min = $time[1];
$hour = sprintf("%02d", $time[2]);
$day = sprintf("%02d", $time[$DAY]);
$month = sprintf("%02d", $time[$MON] + 1); # shift from January=0 to January=1
setyear($time[$YEAR]);
#$year = sprintf("%02d", $time[$YEAR] % 100); # 2-digit year
#$year4 = sprintf("%04d", $time[$YEAR] + 1900); # 4-digit year
$monthdayhour = sprintf("%02d-%02d-%02d", $month, $day, $hour);
$yearmonthday = sprintf("%04d-%02d-%02d", $year4, $month, $day);
$monthday = sprintf("%02d-%02d", $month, $day);
#DEBUG
#print "[$yearmonthday|$monthdayhour|$year4|$year|$month|$day|$hour]\n";

# constant useful in time manipulations
$secs_per_day = 86400;
@month_names=qw(
		start_with_1 January February March April May June
		July August September October November December
);

# Here's the way to get other days:
#@yesterday = localtime(timelocal(@time)-$secs_per_day);


######################################################################
#
#  other non-configuration constants
#
######################################################################

# patterns to match flag strings that indicate suspect min and max
%bad_flag_regex = (
		   'max' => 'Ix|Ex|Px|B|X',
		   'min' => 'In|En|Pn|B|X'
		   );
# backwards compatibility
$pat_min = $bad_flag_regex{'min'};
$pat_max = $bad_flag_regex{'max'};

######################################################################
#
#  utility subroutines
#
######################################################################


BEGIN {
    $_util_log10 = log(10);
    $_util_log2 = log(2);
}
# and now for that other kind of log
sub log10 ( $ ) {
    return log($_[0])/$_util_log10;
}
sub log2 ( $ ) {
    return log($_[0])/$_util_log2;
}

sub dumphash ($%) {
    my ($title, $aa) = @_;

    print "$title\n";
    #print "[", scalar(keys %$aa), "]";
    print "(\n";
    foreach (sort { $a cmp $b} keys %$aa) {
	print "$_ => ",defined($aa->{$_}) ? $aa->{$_} : "__UNDEF", "\n";
    }
    print ");\n";
}

# dump the hash to a string, not to stdout.  The string should evaluate
# to a hash ref; the contents should be just like those of the original
# hash, except that (1) references are broken, and (2) undefined values
# have been mapped to '__UNDEF'.  The latter can be undone by:
# foreach (keys %$f) { undef $f->{$_} if $f->{$_} eq '__UNDEF' }
sub dumphash_str (%) {
    my ($aa) = @_;

    #print "[", scalar(keys %$aa), "]";
    my $retval = '';
    $retval .= "{";
    my $firsttime = 1;
    foreach (sort { $a cmp $b} keys %$aa) {
	$retval .= ',' unless $firsttime;
	$firsttime = 0;
	$retval .= "\n$_ => ";
	$retval .= (defined($aa->{$_}) ? $aa->{$_} : "undef()");
    }
    $retval .= "\n};\n";
    return $retval;
}

# dump the hash, but sort the keys numerically
sub dumphash_num ($%) {
    my ($title, $aa) = @_;

    print "$title\n";
    #print "[", scalar(keys %$aa), "]";
    print "(\n";
    foreach (sort { $a <=> $b} keys %$aa) {
        print "$_ => $aa->{$_}\n";
    }
    print ")\n";
}


# set the global array @days_in_month of the number of days in each month.
# The 0: the array base is 0, but we use jan=1, so use a placeholder in the
# 0th slot.  The point of this subroutine is to deal with leap years.
# It also sets $year and $year4. WARNING: not y2050 compliant.
# input can be '' => no year; historical averages
# [0,99]          => # assumed to be between 1950 and 2049.
# 100-199         => input+1900
# >=200           => absolute (four digit) year
# WARNING: not y2050 compliant
sub setyear($) {
    my($y) = @_;
    if ($y eq "") {
	# refers to historical data; don't set the years themselves.
	$leap = 1;
    } else {
	# setting an actual year -- the normal case
	if ($y<100) {          # two digit year
	    $year4 = 1950 + ($y+50)%100;  # map to 1950-2049 interval
	} elsif ($y<200) {     # standard perl year
	    $year4 = $y+1900;
	} else {               # four-digit year
	    $year4 = $y;
	}
	$year = sprintf("%02d", $year4 % 100); # 2-digit year
	$time[$YEAR] = $year4 - 1900;
	$year4 = sprintf("%04d", $year4);
	$leap = ($year4 % 4 == 0) - ($year4 % 100 == 0) + ($year4 % 400 == 0);
    }
    @days_in_month=(0,31,$leap?29:28,31,30,31,30,31,31,30,31,30,31);
    @month_doy_offsets=(0);
    for my $mon (1..12) {
	$month_doy_offsets[$mon] = 
	    $month_doy_offsets[$mon-1] + $days_in_month[$mon-1];
    }
}


sub is_a_number($) {
    my($str) = @_;
    
    # This check suggested in the perldata man page
    if ($str =~ /^\s*[+-]?(?:\d+(?:\.\d*)?|\.\d+)\s*$/) { return 1 }
    return 0;
}

# convert a numeric month and day to a string with standard formatting
#sub datestr($$) { No prototype because that forces scalar context
sub datestr ($$) {
    my ($month, $day) = @_;
    return sprintf("%2d %2d", $month, $day);
}

# incdate: increment the date given by month and day.
sub incdate (\$\$) {
    my($mon,$day) = @_;

    $$day++;
    if ($$day > $days_in_month[$$mon]) {  # new month
        $$mon++;
        $$day = 1;
        $$mon = 1 if $$mon > 12;
    }
    return $$mon, $$day;
}

# incdate2: increment the date #2 - DONT MODIFY ORIG params - RTN NEW - LEN
sub incdate2 ($$) {
    my($mon,$day) = @_;

    $day++;
    if ($day > $days_in_month[$mon]) {  # new month
        $mon++;
        $mon = 1 if $mon > 12;
        $day = 1;
    }
    return $mon, $day;
}

# increment the date by one day given the month and the day in string form
sub incdate_str($){
    my ($m, $d) = split'[\s-]', $_[0];
    incdate($m,$d);
    return datestr($m,$d);
}

# decdate2: decrement the date #2 - DONT MODIFY ORIG params - RTN NEW - LEN
sub decdate2 ($$) {
    my($mon,$day) = @_;

    $day--;
    if ($day < 1 ) {  # prev month
        $mon--;
        $mon = 12 if $mon < 1;
        $day = $days_in_month[$mon];
    }
    return $mon, $day;
}

# comparing dates
sub cmpdatenum ($$$$) {
    my ($amon, $aday, $bmon, $bday) = @_;
    return ($amon == $bmon) ? $aday <=> $bday : $amon <=> $bmon;
}

sub cmpdatestr {
    my ($a, $b) = @_;
    my ($amon, $aday) = split /[ -]/, $a;
    my ($bmon, $bday) = split /[ -]/, $b;
    return ($amon == $bmon) ? $aday <=> $bday : $amon <=> $bmon;
}

# used with sort(); the args are passed in $a and $b, rather than by the
# usual mechanism.  As usual, we need to use them again to avoid warnings
$a = ""; $b="";
sub cmpdatestr_ab {
    my ($amon, $aday) = split ' ', $a;
    my ($bmon, $bday) = split ' ', $b;
    return ($amon == $bmon) ? $aday <=> $bday : $amon <=> $bmon;
}


# return the max (or min) of two numeric args.
sub max($$) {
    return $_[0] > $_[1] ? $_[0] : $_[1];
}
sub min($$) {
    return $_[0] > $_[1] ? $_[1] : $_[0];
}

# remove leading and trailing whitespace
sub trim_white( $ ) {
    my $str = shift;
    $str =~ s/^\s+//;
    $str =~ s/\s+$//;
    return $str;
}


# evaluate a polynomial in one variable of arbitrary degree
# Pass constant term first, so that $coeffs[$i] goes with $x**$i
sub polynomial ( $@ ) {
    my $x = shift;
    my @coeffs = @_;  # const coeff first; sum $coeffs[$i]*$x**$i

    my $sum = 0;
    foreach (reverse @coeffs) {
	$sum = $sum * $x + $_;
    }
    return $sum;
}

sub hhmm_hrsmins ( $ ) {
    my ($hhmm ) = @_;

    my $hrs = int($hhmm/100);
    my $min = $hhmm - 100*$hrs;

    return $hrs, $min;
}

sub hhmm_hrs ( $ ) {
    my ($hhmm ) = @_;

    return int($hhmm/100);
}

sub hhmm_mins ( $ ) {
    my ($hhmm ) = @_;

    return (hhmm_hrsmins($hhmm))[1];
}

sub hhmm ( $$ ) {
    return 100*$_[0]+$_[1];
}

sub decimal_hrs ( $$ ) {
    return $_[0]+($_[1]/60);
}

# takes to times in hhmm format and returns the difference in decimal hours
sub hhmm_diff( $$ ) {
    my ($arg1, $arg2) = @_;
    my ($h1, $m1) = hhmm_hrsmins($arg1);
    my ($h2, $m2) = hhmm_hrsmins($arg2);

    return decimal_hrs($h1-$h2, $m1-$m2);
}

# assumes @days_in_month is correct for year in question; see setyear()
sub datetodoy( $$ ) {
    my ($month, $day) = @_;

    # Convention allows e.g. 4-33 so that calling code doesn't have to
    # look up the number of days in the month.  This is fixed here.
    if ($day > $days_in_month[$month]) {
	$day = $days_in_month[$month];
    }

    return $month_doy_offsets[$month] + $day;
}

# assumes @days_in_month is correct for year in question; see setyear()
# does not account for leap year
sub datetodoywinter( $$ ) {
    my ($month, $day) = @_;

    # Convention allows e.g. 4-33 so that calling code doesn't have to
    # look up the number of days in the month.  This is fixed here.
    if ($day > $days_in_month[$month]) {
	$day = $days_in_month[$month];
    }

    my $doy = $month_doy_offsets[$month] + $day;
    if ($doy<180) { $doy += datetodoy(12,31);}
    return $doy;
}

# assumes @days_in_month is correct for year in question; see setyear()
sub doytodate( $ ) {
    my $doy = shift;

    # maybe this should be a binary search, but it won't save more 3x
    my $month = 1;
    while($doy > $days_in_month[$month]) {
	$doy -= $days_in_month[$month];
	++$month;
    }
    return $month, $doy;
}


# assumes @days_in_month is correct for year in question; see setyear()
sub doytodatewinter( $ ) {
    my $doy = shift;
    if($doy>datetodoy(12,31)) {
      $doy-= datetodoy(12,31);
    }  
    # maybe this should be a binary search, but it won't save more 3x
    my $month = 1;
    while($doy > $days_in_month[$month]) {
	$doy -= $days_in_month[$month];
	++$month;
    }
    return $month, $doy;
}

# take either a date or a doy, return a doy
sub make_doy( $ ) {
    my $input = shift;

    if ($input =~ /^(\d+)[-\/](\d+)$/) {    # if mm/dd or mm-dd
        return datetodoy($1,$2);     # convert to doy
    }
    return $input;  # else it was a doy
}


# take either a date or a doy, return a doy
sub make_doywinter( $ ) {
    my $input = shift;

    if ($input =~ /^(\d+)[-\/](\d+)$/) {    # if mm/dd or mm-dd
        return datetodoywinter($1,$2);     # convert to doy
    }
    return $input;  # else it was a doy
}

# take either a date or a doy, return a date, i.e. ($m, $d) list
sub make_date( $ ) {
    my $input = shift;

    if ($input =~ /^(\d+)[-\/](\d+)$/) {    # if mm/dd or mm-dd
        return ($1,$2);                     # return ($m, $d)
    }
    return doytodate($input);  # else it was a doy
}


# convert degrees, minutes, seconds to decimal degrees
# e.g. 121:34:40W --> -121.57778
sub dms2dd( $ ) {
    my $in = shift @_;
    my $dir = $in =~ s/([NSEW])$//i  ?  $1  :  '';
    my @s = split ':', $in;
    my $accum = 0;
    while(1) {
	$accum += pop @s;
	last unless @s;
	$accum /= 60;
    }
    $accum *= -1 if $dir =~ /[SW]/i;  # direction convention
    return $accum;
}
# convert decimal degrees to degrees, minutes, seconds, with conventions
# -121.57778, E -> 121:34:40W; 32.8333 -> 32:50:00
sub dd2dms( $;$ ) {
    my ($dd, $axis) = @_;

    my $ret = '00:00:00';
    my $sign = 0;
    if ($dd != 0) {
	$sign = abs($dd)/$dd;
	$dd = abs($dd);

	my $deg = int($dd);
	my $mm = 60*($dd-$deg);
	my $min = int($mm);
	my $sec = 60*($mm-$min);

	# fix some roundoff error
	if ($sec>=59.995) {
	    $sec = 0;
	    $min += 1;
	    if ($min >= 60) {
		$min = 0;
		$seg += 1;
	    }
	}

	$ret = sprintf "%3d:%02d:%05.2f", $deg, $min, $sec;

    }
    if ($axis) {
	if ($axis =~ /[EW]/i) { # it's an easting; use E or W as appropriate
	    $ret .= ( $sign>=0 ? 'E' : 'W' );
	} elsif ($axis =~ /[NS]/i) { # it's a northing; use N or S
	    $ret .= ( $sign>=0 ? 'N' : 'S' );
	} else {
	    die "bad axis $axis";
	}
    } else {
	# we don't have an axis, so just use a '-' if needed
	$ret = '-'.$ret if $sign<0;
    }
    return $ret;
}

# take either dms or dd and return it in dms form
sub dms( $;$ ) {
    return $_[0] =~ /:/  ?  $_[0]  :  dd2dms($_[0],$_[1])
}

# take either dms or dd and return it in dd form
sub dd( $ ) {
    return $_[0] =~ /[:NESW]/  ?  dms2dd($_[0])  :  $_[0]
}

sub round_dms( $ ) {
    my $in = shift @_;
    $in =~ s/([NSEW])$//i;
    my $dir = $1;
    my @s = split ':', $in;
    $s[$#s] = int(0.5 + $s[$#s]);

    # cope with missing seconds or even minutes
    if ($#s<2) {
	$s[1] = 0 if $#s<1;
	$s[2] = 0;
    }


    # we might have rounded 42:25:59.9999 to 42:25:60
    # do carries to map it to 42:26:00
    for($i=$#s; $i>=1; $i--) {
	if ($s[$i] >= 60) {
	    $s[$i] -= 60;
	    $s[$i-1] += 1;
	}
    }

    $ret = sprintf "%d:%02d:%02d%s", $s[0], $s[1], $s[2], $dir; 
}


return 1; # a "require"ed file must return a true value